import datetime
import random
import subprocess
import sys

from OpenSSL import crypto
from bottle import template
from six import string_types

from eu.softfire.tub.utils.utils import get_config, get_logger

KEY_LENGTH_CHOICES = {
    'none': ('', ''),
    '512': ('512', '512'),
    '1024': ('1024', '1024'),
    '2048': ('2048', '2048'),
    '4096': ('4096', '4096')
}

DEFAULT_CERT_VALIDITY = 1
DEFAULT_KEY_LENGTH = '2048'
DEFAULT_DIGEST_ALGORITHM = 'sha256'
CERT_KEYUSAGE_CRITICAL = False
CERT_KEYUSAGE_VALUE = 'digitalSignature, keyEncipherment'
CRL_PROTECTED = False

generalized_time = '%Y%m%d%H%M%SZ'
logger = get_logger(__name__)


def bytes_compat(string, encoding='utf-8'):
    if sys.version_info.major >= 3:
        if not isinstance(string, string_types):
            string = str(string)
        return bytes(string, encoding)
    else:
        return bytes(string)


class CertificateGenerator(object):
    def __init__(self):
        self.key_length = 2048
        self.digest = 'sha1'
        with open(get_config('system', 'cert-ca-file', '/etc/softfire/softfire-ca.p12'), 'rb') as buf:
            buffer = buf.read()
        load_pkcs_ = crypto.load_pkcs12(buffer=buffer,
                                        passphrase=get_config('system', 'cert-passphrase', 'softfire').encode())
        self.ca_cert = load_pkcs_.get_certificate()
        self.ca_key = load_pkcs_.get_privatekey()
        self.certificate = None
        self.private_key = None
        with open(get_config('system', 'openvpn-template-location', '/etc/softfire/template_openvpn.tpl'), 'r') as f:
            self.openvpn_config_tpl = f.read()

    def generate(self, passphrase: str = None, common_name=None, days=DEFAULT_CERT_VALIDITY):
        k = crypto.PKey()
        k.generate_key(crypto.TYPE_RSA, self.key_length)

        cert = crypto.X509()
        # cert.get_subject().CN = common_name
        cert.get_subject().commonName = common_name
        cert.set_serial_number(random.randint(990000, 999999999999999999999999999))
        cert.gmtime_adj_notBefore(-600)
        cert.gmtime_adj_notAfter(int(datetime.timedelta(days=days).total_seconds()))
        cert.set_issuer(self.ca_cert.get_subject())
        cert.set_pubkey(k)
        cert = self._add_extensions(cert)
        cert.sign(self.ca_key, self.digest)

        self.certificate = crypto.dump_certificate(crypto.FILETYPE_PEM, cert)
        self.private_key = crypto.dump_privatekey(crypto.FILETYPE_PEM, k, cipher="DES-EDE3-CBC", passphrase=passphrase.encode())
        return self

    def _add_extensions(self, cert):
        """
        (internal use only)
        adds x509 extensions to ``cert``
        """
        ext = list()
        ext.append(crypto.X509Extension(b'basicConstraints',
                                        True,
                                        b'CA:FALSE'))
        ext.append(crypto.X509Extension(b'keyUsage',
                                        CERT_KEYUSAGE_CRITICAL,
                                        bytes_compat(CERT_KEYUSAGE_VALUE)))
        issuer_cert = self.ca_cert
        ext.append(crypto.X509Extension(b'subjectKeyIdentifier',
                                        False,
                                        b'hash',
                                        subject=cert))
        cert.add_extensions(ext)
        cert.add_extensions([
            crypto.X509Extension(b'authorityKeyIdentifier',
                                 False,
                                 b'keyid:always,issuer:always',
                                 issuer=issuer_cert)
        ])
        return cert

    def get_openvpn_config(self):
        openvpn_options = {
            'openvpn_server': get_config('openvpn', 'openvpn_server', 'softfire-vpn.av.tu-berlin.de'),
            'openvpn_port': int(get_config('openvpn', 'openvpn_port', "443")),
            'protocol': get_config('openvpn', 'openvpn_protocol', 'tcp'),
            'certificate': self.certificate.decode("utf-8"), 'key': self.private_key.decode("utf-8")
        }
        return template(self.openvpn_config_tpl, openvpn_options)


if __name__ == '__main__':
    cert_gen = CertificateGenerator()
    cert_gen.generate(passphrase='123456', common_name="foobar", days=1)
    print(cert_gen.get_openvpn_config())
    print("")
    print("")
    print("")
    print("")

    output = subprocess.check_output(['openssl', 'x509', '-noout', '-text'], input=cert_gen.certificate).decode('utf-8')
    print(output)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
import grpc

import eu.softfire.tub.messaging.grpc.messages_pb2 as messages__pb2


class RegistrationServiceStub(object):

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.register = channel.unary_unary(
        '/RegistrationService/register',
        request_serializer=messages__pb2.RegisterMessage.SerializeToString,
        response_deserializer=messages__pb2.ResponseMessage.FromString,
        )
    self.unregister = channel.unary_unary(
        '/RegistrationService/unregister',
        request_serializer=messages__pb2.UnregisterMessage.SerializeToString,
        response_deserializer=messages__pb2.ResponseMessage.FromString,
        )
    self.update_status = channel.unary_unary(
        '/RegistrationService/update_status',
        request_serializer=messages__pb2.StatusMessage.SerializeToString,
        response_deserializer=messages__pb2.ResponseMessage.FromString,
        )


class RegistrationServiceServicer(object):

  def register(self, request, context):
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def unregister(self, request, context):
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def update_status(self, request, context):
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_RegistrationServiceServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'register': grpc.unary_unary_rpc_method_handler(
          servicer.register,
          request_deserializer=messages__pb2.RegisterMessage.FromString,
          response_serializer=messages__pb2.ResponseMessage.SerializeToString,
      ),
      'unregister': grpc.unary_unary_rpc_method_handler(
          servicer.unregister,
          request_deserializer=messages__pb2.UnregisterMessage.FromString,
          response_serializer=messages__pb2.ResponseMessage.SerializeToString,
      ),
      'update_status': grpc.unary_unary_rpc_method_handler(
          servicer.update_status,
          request_deserializer=messages__pb2.StatusMessage.FromString,
          response_serializer=messages__pb2.ResponseMessage.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'RegistrationService', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))


class ManagerAgentStub(object):

  def __init__(self, channel):
    """Constructor.

    Args:
      channel: A grpc.Channel.
    """
    self.execute = channel.unary_unary(
        '/ManagerAgent/execute',
        request_serializer=messages__pb2.RequestMessage.SerializeToString,
        response_deserializer=messages__pb2.ResponseMessage.FromString,
        )
    self.refresh_resources = channel.unary_unary(
        '/ManagerAgent/refresh_resources',
        request_serializer=messages__pb2.UserInfo.SerializeToString,
        response_deserializer=messages__pb2.ResponseMessage.FromString,
        )
    self.create_user = channel.unary_unary(
        '/ManagerAgent/create_user',
        request_serializer=messages__pb2.UserInfo.SerializeToString,
        response_deserializer=messages__pb2.UserInfo.FromString,
        )


class ManagerAgentServicer(object):

  def execute(self, request, context):
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def refresh_resources(self, request, context):
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')

  def create_user(self, request, context):
    context.set_code(grpc.StatusCode.UNIMPLEMENTED)
    context.set_details('Method not implemented!')
    raise NotImplementedError('Method not implemented!')


def add_ManagerAgentServicer_to_server(servicer, server):
  rpc_method_handlers = {
      'execute': grpc.unary_unary_rpc_method_handler(
          servicer.execute,
          request_deserializer=messages__pb2.RequestMessage.FromString,
          response_serializer=messages__pb2.ResponseMessage.SerializeToString,
      ),
      'refresh_resources': grpc.unary_unary_rpc_method_handler(
          servicer.refresh_resources,
          request_deserializer=messages__pb2.UserInfo.FromString,
          response_serializer=messages__pb2.ResponseMessage.SerializeToString,
      ),
      'create_user': grpc.unary_unary_rpc_method_handler(
          servicer.create_user,
          request_deserializer=messages__pb2.UserInfo.FromString,
          response_serializer=messages__pb2.UserInfo.SerializeToString,
      ),
  }
  generic_handler = grpc.method_handlers_generic_handler(
      'ManagerAgent', rpc_method_handlers)
  server.add_generic_rpc_handlers((generic_handler,))



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: messages.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='messages.proto',
  package='',
  syntax='proto3',
  serialized_pb=_b('\n\x0emessages.proto\"F\n\x0fRegisterMessage\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08\x65ndpoint\x18\x02 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\"3\n\x11UnregisterMessage\x12\x0c\n\x04name\x18\x01 \x01(\t\x12\x10\n\x08\x65ndpoint\x18\x02 \x01(\t\"U\n\rStatusMessage\x12\x1c\n\tresources\x18\x01 \x03(\x0b\x32\t.Resource\x12\x10\n\x08username\x18\x02 \x01(\t\x12\x14\n\x0cmanager_name\x18\x03 \x01(\t\"X\n\x0eRequestMessage\x12\x17\n\x06method\x18\x01 \x01(\x0e\x32\x07.Method\x12\x0f\n\x07payload\x18\x02 \x01(\t\x12\x1c\n\tuser_info\x18\x03 \x01(\x0b\x32\t.UserInfo\"\xe8\x01\n\x0fResponseMessage\x12\x17\n\x06result\x18\x01 \x01(\x0e\x32\x07.Result\x12.\n\rlist_resource\x18\x02 \x01(\x0b\x32\x15.ListResourceResponseH\x00\x12\x34\n\x10provide_resource\x18\x03 \x01(\x0b\x32\x18.ProvideResourceResponseH\x00\x12\x34\n\x10refresh_resource\x18\x04 \x01(\x0b\x32\x18.RefreshResourceResponseH\x00\x12\x15\n\rerror_message\x18\x05 \x01(\tB\t\n\x07message\"<\n\x14ListResourceResponse\x12$\n\tresources\x18\x01 \x03(\x0b\x32\x11.ResourceMetadata\"7\n\x17ProvideResourceResponse\x12\x1c\n\tresources\x18\x01 \x03(\x0b\x32\t.Resource\"?\n\x17RefreshResourceResponse\x12$\n\tresources\x18\x01 \x03(\x0b\x32\x11.ResourceMetadata\"\x7f\n\x10ResourceMetadata\x12\x13\n\x0bresource_id\x18\x01 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x02 \x01(\t\x12\x13\n\x0b\x63\x61rdinality\x18\x03 \x01(\x05\x12\x19\n\x07testbed\x18\x04 \x01(\x0e\x32\x08.Testbed\x12\x11\n\tnode_type\x18\x05 \x01(\t\"\xbc\x01\n\x08UserInfo\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x10\n\x08password\x18\x03 \x01(\t\x12\x15\n\rob_project_id\x18\x04 \x01(\t\x12\x36\n\x0ftestbed_tenants\x18\x05 \x03(\x0b\x32\x1d.UserInfo.TestbedTenantsEntry\x1a\x35\n\x13TestbedTenantsEntry\x12\x0b\n\x03key\x18\x01 \x01(\x05\x12\r\n\x05value\x18\x02 \x01(\t:\x02\x38\x01\"\'\n\x08Resource\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0f\n\x07\x63ontent\x18\x02 \x01(\t*b\n\x06Method\x12\x12\n\x0eLIST_RESOURCES\x10\x00\x12\x15\n\x11PROVIDE_RESOURCES\x10\x01\x12\x15\n\x11RELEASE_RESOURCES\x10\x02\x12\x16\n\x12VALIDATE_RESOURCES\x10\x03*\x1b\n\x06Result\x12\x06\n\x02Ok\x10\x00\x12\t\n\x05\x45RROR\x10\x01*\x92\x01\n\x07Testbed\x12\n\n\x06SURREY\x10\x00\x12\t\n\x05\x46OKUS\x10\x01\x12\x06\n\x02\x44T\x10\x02\x12\x07\n\x03\x41\x44S\x10\x03\x12\x0c\n\x08\x45RICSSON\x10\x04\x12\x0e\n\nSURREY_DEV\x10\x05\x12\r\n\tFOKUS_DEV\x10\x06\x12\n\n\x06\x44T_DEV\x10\x07\x12\x0b\n\x07\x41\x44S_DEV\x10\x08\x12\x10\n\x0c\x45RICSSON_DEV\x10\t\x12\x07\n\x03\x41NY\x10\n2\xb2\x01\n\x13RegistrationService\x12\x30\n\x08register\x12\x10.RegisterMessage\x1a\x10.ResponseMessage\"\x00\x12\x34\n\nunregister\x12\x12.UnregisterMessage\x1a\x10.ResponseMessage\"\x00\x12\x33\n\rupdate_status\x12\x0e.StatusMessage\x1a\x10.ResponseMessage\"\x00\x32\x99\x01\n\x0cManagerAgent\x12.\n\x07\x65xecute\x12\x0f.RequestMessage\x1a\x10.ResponseMessage\"\x00\x12\x32\n\x11refresh_resources\x12\t.UserInfo\x1a\x10.ResponseMessage\"\x00\x12%\n\x0b\x63reate_user\x12\t.UserInfo\x1a\t.UserInfo\"\x00\x42\x02H\x03\x62\x06proto3')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

_METHOD = _descriptor.EnumDescriptor(
  name='Method',
  full_name='Method',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='LIST_RESOURCES', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='PROVIDE_RESOURCES', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='RELEASE_RESOURCES', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='VALIDATE_RESOURCES', index=3, number=3,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=1100,
  serialized_end=1198,
)
_sym_db.RegisterEnumDescriptor(_METHOD)

Method = enum_type_wrapper.EnumTypeWrapper(_METHOD)
_RESULT = _descriptor.EnumDescriptor(
  name='Result',
  full_name='Result',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='Ok', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ERROR', index=1, number=1,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=1200,
  serialized_end=1227,
)
_sym_db.RegisterEnumDescriptor(_RESULT)

Result = enum_type_wrapper.EnumTypeWrapper(_RESULT)
_TESTBED = _descriptor.EnumDescriptor(
  name='Testbed',
  full_name='Testbed',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='SURREY', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FOKUS', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ADS', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ERICSSON', index=4, number=4,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='SURREY_DEV', index=5, number=5,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FOKUS_DEV', index=6, number=6,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DT_DEV', index=7, number=7,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ADS_DEV', index=8, number=8,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ERICSSON_DEV', index=9, number=9,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='ANY', index=10, number=10,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=1230,
  serialized_end=1376,
)
_sym_db.RegisterEnumDescriptor(_TESTBED)

Testbed = enum_type_wrapper.EnumTypeWrapper(_TESTBED)
LIST_RESOURCES = 0
PROVIDE_RESOURCES = 1
RELEASE_RESOURCES = 2
VALIDATE_RESOURCES = 3
Ok = 0
ERROR = 1
SURREY = 0
FOKUS = 1
DT = 2
ADS = 3
ERICSSON = 4
SURREY_DEV = 5
FOKUS_DEV = 6
DT_DEV = 7
ADS_DEV = 8
ERICSSON_DEV = 9
ANY = 10



_REGISTERMESSAGE = _descriptor.Descriptor(
  name='RegisterMessage',
  full_name='RegisterMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='RegisterMessage.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='endpoint', full_name='RegisterMessage.endpoint', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='description', full_name='RegisterMessage.description', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=18,
  serialized_end=88,
)


_UNREGISTERMESSAGE = _descriptor.Descriptor(
  name='UnregisterMessage',
  full_name='UnregisterMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='name', full_name='UnregisterMessage.name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='endpoint', full_name='UnregisterMessage.endpoint', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=90,
  serialized_end=141,
)


_STATUSMESSAGE = _descriptor.Descriptor(
  name='StatusMessage',
  full_name='StatusMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='resources', full_name='StatusMessage.resources', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='username', full_name='StatusMessage.username', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='manager_name', full_name='StatusMessage.manager_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=143,
  serialized_end=228,
)


_REQUESTMESSAGE = _descriptor.Descriptor(
  name='RequestMessage',
  full_name='RequestMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='method', full_name='RequestMessage.method', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='payload', full_name='RequestMessage.payload', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='user_info', full_name='RequestMessage.user_info', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=230,
  serialized_end=318,
)


_RESPONSEMESSAGE = _descriptor.Descriptor(
  name='ResponseMessage',
  full_name='ResponseMessage',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='result', full_name='ResponseMessage.result', index=0,
      number=1, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='list_resource', full_name='ResponseMessage.list_resource', index=1,
      number=2, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='provide_resource', full_name='ResponseMessage.provide_resource', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='refresh_resource', full_name='ResponseMessage.refresh_resource', index=3,
      number=4, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='error_message', full_name='ResponseMessage.error_message', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='message', full_name='ResponseMessage.message',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=321,
  serialized_end=553,
)


_LISTRESOURCERESPONSE = _descriptor.Descriptor(
  name='ListResourceResponse',
  full_name='ListResourceResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='resources', full_name='ListResourceResponse.resources', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=555,
  serialized_end=615,
)


_PROVIDERESOURCERESPONSE = _descriptor.Descriptor(
  name='ProvideResourceResponse',
  full_name='ProvideResourceResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='resources', full_name='ProvideResourceResponse.resources', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=617,
  serialized_end=672,
)


_REFRESHRESOURCERESPONSE = _descriptor.Descriptor(
  name='RefreshResourceResponse',
  full_name='RefreshResourceResponse',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='resources', full_name='RefreshResourceResponse.resources', index=0,
      number=1, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=674,
  serialized_end=737,
)


_RESOURCEMETADATA = _descriptor.Descriptor(
  name='ResourceMetadata',
  full_name='ResourceMetadata',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='resource_id', full_name='ResourceMetadata.resource_id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='description', full_name='ResourceMetadata.description', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='cardinality', full_name='ResourceMetadata.cardinality', index=2,
      number=3, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='testbed', full_name='ResourceMetadata.testbed', index=3,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='node_type', full_name='ResourceMetadata.node_type', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=739,
  serialized_end=866,
)


_USERINFO_TESTBEDTENANTSENTRY = _descriptor.Descriptor(
  name='TestbedTenantsEntry',
  full_name='UserInfo.TestbedTenantsEntry',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='key', full_name='UserInfo.TestbedTenantsEntry.key', index=0,
      number=1, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='value', full_name='UserInfo.TestbedTenantsEntry.value', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=_descriptor._ParseOptions(descriptor_pb2.MessageOptions(), _b('8\001')),
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1004,
  serialized_end=1057,
)

_USERINFO = _descriptor.Descriptor(
  name='UserInfo',
  full_name='UserInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='UserInfo.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='name', full_name='UserInfo.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='password', full_name='UserInfo.password', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='ob_project_id', full_name='UserInfo.ob_project_id', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='testbed_tenants', full_name='UserInfo.testbed_tenants', index=4,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_USERINFO_TESTBEDTENANTSENTRY, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=869,
  serialized_end=1057,
)


_RESOURCE = _descriptor.Descriptor(
  name='Resource',
  full_name='Resource',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='Resource.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='content', full_name='Resource.content', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=1059,
  serialized_end=1098,
)

_STATUSMESSAGE.fields_by_name['resources'].message_type = _RESOURCE
_REQUESTMESSAGE.fields_by_name['method'].enum_type = _METHOD
_REQUESTMESSAGE.fields_by_name['user_info'].message_type = _USERINFO
_RESPONSEMESSAGE.fields_by_name['result'].enum_type = _RESULT
_RESPONSEMESSAGE.fields_by_name['list_resource'].message_type = _LISTRESOURCERESPONSE
_RESPONSEMESSAGE.fields_by_name['provide_resource'].message_type = _PROVIDERESOURCERESPONSE
_RESPONSEMESSAGE.fields_by_name['refresh_resource'].message_type = _REFRESHRESOURCERESPONSE
_RESPONSEMESSAGE.oneofs_by_name['message'].fields.append(
  _RESPONSEMESSAGE.fields_by_name['list_resource'])
_RESPONSEMESSAGE.fields_by_name['list_resource'].containing_oneof = _RESPONSEMESSAGE.oneofs_by_name['message']
_RESPONSEMESSAGE.oneofs_by_name['message'].fields.append(
  _RESPONSEMESSAGE.fields_by_name['provide_resource'])
_RESPONSEMESSAGE.fields_by_name['provide_resource'].containing_oneof = _RESPONSEMESSAGE.oneofs_by_name['message']
_RESPONSEMESSAGE.oneofs_by_name['message'].fields.append(
  _RESPONSEMESSAGE.fields_by_name['refresh_resource'])
_RESPONSEMESSAGE.fields_by_name['refresh_resource'].containing_oneof = _RESPONSEMESSAGE.oneofs_by_name['message']
_LISTRESOURCERESPONSE.fields_by_name['resources'].message_type = _RESOURCEMETADATA
_PROVIDERESOURCERESPONSE.fields_by_name['resources'].message_type = _RESOURCE
_REFRESHRESOURCERESPONSE.fields_by_name['resources'].message_type = _RESOURCEMETADATA
_RESOURCEMETADATA.fields_by_name['testbed'].enum_type = _TESTBED
_USERINFO_TESTBEDTENANTSENTRY.containing_type = _USERINFO
_USERINFO.fields_by_name['testbed_tenants'].message_type = _USERINFO_TESTBEDTENANTSENTRY
DESCRIPTOR.message_types_by_name['RegisterMessage'] = _REGISTERMESSAGE
DESCRIPTOR.message_types_by_name['UnregisterMessage'] = _UNREGISTERMESSAGE
DESCRIPTOR.message_types_by_name['StatusMessage'] = _STATUSMESSAGE
DESCRIPTOR.message_types_by_name['RequestMessage'] = _REQUESTMESSAGE
DESCRIPTOR.message_types_by_name['ResponseMessage'] = _RESPONSEMESSAGE
DESCRIPTOR.message_types_by_name['ListResourceResponse'] = _LISTRESOURCERESPONSE
DESCRIPTOR.message_types_by_name['ProvideResourceResponse'] = _PROVIDERESOURCERESPONSE
DESCRIPTOR.message_types_by_name['RefreshResourceResponse'] = _REFRESHRESOURCERESPONSE
DESCRIPTOR.message_types_by_name['ResourceMetadata'] = _RESOURCEMETADATA
DESCRIPTOR.message_types_by_name['UserInfo'] = _USERINFO
DESCRIPTOR.message_types_by_name['Resource'] = _RESOURCE
DESCRIPTOR.enum_types_by_name['Method'] = _METHOD
DESCRIPTOR.enum_types_by_name['Result'] = _RESULT
DESCRIPTOR.enum_types_by_name['Testbed'] = _TESTBED

RegisterMessage = _reflection.GeneratedProtocolMessageType('RegisterMessage', (_message.Message,), dict(
  DESCRIPTOR = _REGISTERMESSAGE,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:RegisterMessage)
  ))
_sym_db.RegisterMessage(RegisterMessage)

UnregisterMessage = _reflection.GeneratedProtocolMessageType('UnregisterMessage', (_message.Message,), dict(
  DESCRIPTOR = _UNREGISTERMESSAGE,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:UnregisterMessage)
  ))
_sym_db.RegisterMessage(UnregisterMessage)

StatusMessage = _reflection.GeneratedProtocolMessageType('StatusMessage', (_message.Message,), dict(
  DESCRIPTOR = _STATUSMESSAGE,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:StatusMessage)
  ))
_sym_db.RegisterMessage(StatusMessage)

RequestMessage = _reflection.GeneratedProtocolMessageType('RequestMessage', (_message.Message,), dict(
  DESCRIPTOR = _REQUESTMESSAGE,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:RequestMessage)
  ))
_sym_db.RegisterMessage(RequestMessage)

ResponseMessage = _reflection.GeneratedProtocolMessageType('ResponseMessage', (_message.Message,), dict(
  DESCRIPTOR = _RESPONSEMESSAGE,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:ResponseMessage)
  ))
_sym_db.RegisterMessage(ResponseMessage)

ListResourceResponse = _reflection.GeneratedProtocolMessageType('ListResourceResponse', (_message.Message,), dict(
  DESCRIPTOR = _LISTRESOURCERESPONSE,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:ListResourceResponse)
  ))
_sym_db.RegisterMessage(ListResourceResponse)

ProvideResourceResponse = _reflection.GeneratedProtocolMessageType('ProvideResourceResponse', (_message.Message,), dict(
  DESCRIPTOR = _PROVIDERESOURCERESPONSE,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:ProvideResourceResponse)
  ))
_sym_db.RegisterMessage(ProvideResourceResponse)

RefreshResourceResponse = _reflection.GeneratedProtocolMessageType('RefreshResourceResponse', (_message.Message,), dict(
  DESCRIPTOR = _REFRESHRESOURCERESPONSE,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:RefreshResourceResponse)
  ))
_sym_db.RegisterMessage(RefreshResourceResponse)

ResourceMetadata = _reflection.GeneratedProtocolMessageType('ResourceMetadata', (_message.Message,), dict(
  DESCRIPTOR = _RESOURCEMETADATA,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:ResourceMetadata)
  ))
_sym_db.RegisterMessage(ResourceMetadata)

UserInfo = _reflection.GeneratedProtocolMessageType('UserInfo', (_message.Message,), dict(

  TestbedTenantsEntry = _reflection.GeneratedProtocolMessageType('TestbedTenantsEntry', (_message.Message,), dict(
    DESCRIPTOR = _USERINFO_TESTBEDTENANTSENTRY,
    __module__ = 'messages_pb2'
    # @@protoc_insertion_point(class_scope:UserInfo.TestbedTenantsEntry)
    ))
  ,
  DESCRIPTOR = _USERINFO,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:UserInfo)
  ))
_sym_db.RegisterMessage(UserInfo)
_sym_db.RegisterMessage(UserInfo.TestbedTenantsEntry)

Resource = _reflection.GeneratedProtocolMessageType('Resource', (_message.Message,), dict(
  DESCRIPTOR = _RESOURCE,
  __module__ = 'messages_pb2'
  # @@protoc_insertion_point(class_scope:Resource)
  ))
_sym_db.RegisterMessage(Resource)


DESCRIPTOR.has_options = True
DESCRIPTOR._options = _descriptor._ParseOptions(descriptor_pb2.FileOptions(), _b('H\003'))
_USERINFO_TESTBEDTENANTSENTRY.has_options = True
_USERINFO_TESTBEDTENANTSENTRY._options = _descriptor._ParseOptions(descriptor_pb2.MessageOptions(), _b('8\001'))
try:
  # THESE ELEMENTS WILL BE DEPRECATED.
  # Please use the generated *_pb2_grpc.py files instead.
  import grpc
  from grpc.beta import implementations as beta_implementations
  from grpc.beta import interfaces as beta_interfaces
  from grpc.framework.common import cardinality
  from grpc.framework.interfaces.face import utilities as face_utilities


  class RegistrationServiceStub(object):

    def __init__(self, channel):
      """Constructor.

      Args:
        channel: A grpc.Channel.
      """
      self.register = channel.unary_unary(
          '/RegistrationService/register',
          request_serializer=RegisterMessage.SerializeToString,
          response_deserializer=ResponseMessage.FromString,
          )
      self.unregister = channel.unary_unary(
          '/RegistrationService/unregister',
          request_serializer=UnregisterMessage.SerializeToString,
          response_deserializer=ResponseMessage.FromString,
          )
      self.update_status = channel.unary_unary(
          '/RegistrationService/update_status',
          request_serializer=StatusMessage.SerializeToString,
          response_deserializer=ResponseMessage.FromString,
          )


  class RegistrationServiceServicer(object):

    def register(self, request, context):
      context.set_code(grpc.StatusCode.UNIMPLEMENTED)
      context.set_details('Method not implemented!')
      raise NotImplementedError('Method not implemented!')

    def unregister(self, request, context):
      context.set_code(grpc.StatusCode.UNIMPLEMENTED)
      context.set_details('Method not implemented!')
      raise NotImplementedError('Method not implemented!')

    def update_status(self, request, context):
      context.set_code(grpc.StatusCode.UNIMPLEMENTED)
      context.set_details('Method not implemented!')
      raise NotImplementedError('Method not implemented!')


  def add_RegistrationServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        'register': grpc.unary_unary_rpc_method_handler(
            servicer.register,
            request_deserializer=RegisterMessage.FromString,
            response_serializer=ResponseMessage.SerializeToString,
        ),
        'unregister': grpc.unary_unary_rpc_method_handler(
            servicer.unregister,
            request_deserializer=UnregisterMessage.FromString,
            response_serializer=ResponseMessage.SerializeToString,
        ),
        'update_status': grpc.unary_unary_rpc_method_handler(
            servicer.update_status,
            request_deserializer=StatusMessage.FromString,
            response_serializer=ResponseMessage.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        'RegistrationService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


  class ManagerAgentStub(object):

    def __init__(self, channel):
      """Constructor.

      Args:
        channel: A grpc.Channel.
      """
      self.execute = channel.unary_unary(
          '/ManagerAgent/execute',
          request_serializer=RequestMessage.SerializeToString,
          response_deserializer=ResponseMessage.FromString,
          )
      self.refresh_resources = channel.unary_unary(
          '/ManagerAgent/refresh_resources',
          request_serializer=UserInfo.SerializeToString,
          response_deserializer=ResponseMessage.FromString,
          )
      self.create_user = channel.unary_unary(
          '/ManagerAgent/create_user',
          request_serializer=UserInfo.SerializeToString,
          response_deserializer=UserInfo.FromString,
          )


  class ManagerAgentServicer(object):

    def execute(self, request, context):
      context.set_code(grpc.StatusCode.UNIMPLEMENTED)
      context.set_details('Method not implemented!')
      raise NotImplementedError('Method not implemented!')

    def refresh_resources(self, request, context):
      context.set_code(grpc.StatusCode.UNIMPLEMENTED)
      context.set_details('Method not implemented!')
      raise NotImplementedError('Method not implemented!')

    def create_user(self, request, context):
      context.set_code(grpc.StatusCode.UNIMPLEMENTED)
      context.set_details('Method not implemented!')
      raise NotImplementedError('Method not implemented!')


  def add_ManagerAgentServicer_to_server(servicer, server):
    rpc_method_handlers = {
        'execute': grpc.unary_unary_rpc_method_handler(
            servicer.execute,
            request_deserializer=RequestMessage.FromString,
            response_serializer=ResponseMessage.SerializeToString,
        ),
        'refresh_resources': grpc.unary_unary_rpc_method_handler(
            servicer.refresh_resources,
            request_deserializer=UserInfo.FromString,
            response_serializer=ResponseMessage.SerializeToString,
        ),
        'create_user': grpc.unary_unary_rpc_method_handler(
            servicer.create_user,
            request_deserializer=UserInfo.FromString,
            response_serializer=UserInfo.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        'ManagerAgent', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


  class BetaRegistrationServiceServicer(object):
    """The Beta API is deprecated for 0.15.0 and later.

    It is recommended to use the GA API (classes and functions in this
    file not marked beta) for all further purposes. This class was generated
    only to ease transition from grpcio<0.15.0 to grpcio>=0.15.0."""
    def register(self, request, context):
      context.code(beta_interfaces.StatusCode.UNIMPLEMENTED)
    def unregister(self, request, context):
      context.code(beta_interfaces.StatusCode.UNIMPLEMENTED)
    def update_status(self, request, context):
      context.code(beta_interfaces.StatusCode.UNIMPLEMENTED)


  class BetaRegistrationServiceStub(object):
    """The Beta API is deprecated for 0.15.0 and later.

    It is recommended to use the GA API (classes and functions in this
    file not marked beta) for all further purposes. This class was generated
    only to ease transition from grpcio<0.15.0 to grpcio>=0.15.0."""
    def register(self, request, timeout, metadata=None, with_call=False, protocol_options=None):
      raise NotImplementedError()
    register.future = None
    def unregister(self, request, timeout, metadata=None, with_call=False, protocol_options=None):
      raise NotImplementedError()
    unregister.future = None
    def update_status(self, request, timeout, metadata=None, with_call=False, protocol_options=None):
      raise NotImplementedError()
    update_status.future = None


  def beta_create_RegistrationService_server(servicer, pool=None, pool_size=None, default_timeout=None, maximum_timeout=None):
    """The Beta API is deprecated for 0.15.0 and later.

    It is recommended to use the GA API (classes and functions in this
    file not marked beta) for all further purposes. This function was
    generated only to ease transition from grpcio<0.15.0 to grpcio>=0.15.0"""
    request_deserializers = {
      ('RegistrationService', 'register'): RegisterMessage.FromString,
      ('RegistrationService', 'unregister'): UnregisterMessage.FromString,
      ('RegistrationService', 'update_status'): StatusMessage.FromString,
    }
    response_serializers = {
      ('RegistrationService', 'register'): ResponseMessage.SerializeToString,
      ('RegistrationService', 'unregister'): ResponseMessage.SerializeToString,
      ('RegistrationService', 'update_status'): ResponseMessage.SerializeToString,
    }
    method_implementations = {
      ('RegistrationService', 'register'): face_utilities.unary_unary_inline(servicer.register),
      ('RegistrationService', 'unregister'): face_utilities.unary_unary_inline(servicer.unregister),
      ('RegistrationService', 'update_status'): face_utilities.unary_unary_inline(servicer.update_status),
    }
    server_options = beta_implementations.server_options(request_deserializers=request_deserializers, response_serializers=response_serializers, thread_pool=pool, thread_pool_size=pool_size, default_timeout=default_timeout, maximum_timeout=maximum_timeout)
    return beta_implementations.server(method_implementations, options=server_options)


  def beta_create_RegistrationService_stub(channel, host=None, metadata_transformer=None, pool=None, pool_size=None):
    """The Beta API is deprecated for 0.15.0 and later.

    It is recommended to use the GA API (classes and functions in this
    file not marked beta) for all further purposes. This function was
    generated only to ease transition from grpcio<0.15.0 to grpcio>=0.15.0"""
    request_serializers = {
      ('RegistrationService', 'register'): RegisterMessage.SerializeToString,
      ('RegistrationService', 'unregister'): UnregisterMessage.SerializeToString,
      ('RegistrationService', 'update_status'): StatusMessage.SerializeToString,
    }
    response_deserializers = {
      ('RegistrationService', 'register'): ResponseMessage.FromString,
      ('RegistrationService', 'unregister'): ResponseMessage.FromString,
      ('RegistrationService', 'update_status'): ResponseMessage.FromString,
    }
    cardinalities = {
      'register': cardinality.Cardinality.UNARY_UNARY,
      'unregister': cardinality.Cardinality.UNARY_UNARY,
      'update_status': cardinality.Cardinality.UNARY_UNARY,
    }
    stub_options = beta_implementations.stub_options(host=host, metadata_transformer=metadata_transformer, request_serializers=request_serializers, response_deserializers=response_deserializers, thread_pool=pool, thread_pool_size=pool_size)
    return beta_implementations.dynamic_stub(channel, 'RegistrationService', cardinalities, options=stub_options)


  class BetaManagerAgentServicer(object):
    """The Beta API is deprecated for 0.15.0 and later.

    It is recommended to use the GA API (classes and functions in this
    file not marked beta) for all further purposes. This class was generated
    only to ease transition from grpcio<0.15.0 to grpcio>=0.15.0."""
    def execute(self, request, context):
      context.code(beta_interfaces.StatusCode.UNIMPLEMENTED)
    def refresh_resources(self, request, context):
      context.code(beta_interfaces.StatusCode.UNIMPLEMENTED)
    def create_user(self, request, context):
      context.code(beta_interfaces.StatusCode.UNIMPLEMENTED)


  class BetaManagerAgentStub(object):
    """The Beta API is deprecated for 0.15.0 and later.

    It is recommended to use the GA API (classes and functions in this
    file not marked beta) for all further purposes. This class was generated
    only to ease transition from grpcio<0.15.0 to grpcio>=0.15.0."""
    def execute(self, request, timeout, metadata=None, with_call=False, protocol_options=None):
      raise NotImplementedError()
    execute.future = None
    def refresh_resources(self, request, timeout, metadata=None, with_call=False, protocol_options=None):
      raise NotImplementedError()
    refresh_resources.future = None
    def create_user(self, request, timeout, metadata=None, with_call=False, protocol_options=None):
      raise NotImplementedError()
    create_user.future = None


  def beta_create_ManagerAgent_server(servicer, pool=None, pool_size=None, default_timeout=None, maximum_timeout=None):
    """The Beta API is deprecated for 0.15.0 and later.

    It is recommended to use the GA API (classes and functions in this
    file not marked beta) for all further purposes. This function was
    generated only to ease transition from grpcio<0.15.0 to grpcio>=0.15.0"""
    request_deserializers = {
      ('ManagerAgent', 'create_user'): UserInfo.FromString,
      ('ManagerAgent', 'execute'): RequestMessage.FromString,
      ('ManagerAgent', 'refresh_resources'): UserInfo.FromString,
    }
    response_serializers = {
      ('ManagerAgent', 'create_user'): UserInfo.SerializeToString,
      ('ManagerAgent', 'execute'): ResponseMessage.SerializeToString,
      ('ManagerAgent', 'refresh_resources'): ResponseMessage.SerializeToString,
    }
    method_implementations = {
      ('ManagerAgent', 'create_user'): face_utilities.unary_unary_inline(servicer.create_user),
      ('ManagerAgent', 'execute'): face_utilities.unary_unary_inline(servicer.execute),
      ('ManagerAgent', 'refresh_resources'): face_utilities.unary_unary_inline(servicer.refresh_resources),
    }
    server_options = beta_implementations.server_options(request_deserializers=request_deserializers, response_serializers=response_serializers, thread_pool=pool, thread_pool_size=pool_size, default_timeout=default_timeout, maximum_timeout=maximum_timeout)
    return beta_implementations.server(method_implementations, options=server_options)


  def beta_create_ManagerAgent_stub(channel, host=None, metadata_transformer=None, pool=None, pool_size=None):
    """The Beta API is deprecated for 0.15.0 and later.

    It is recommended to use the GA API (classes and functions in this
    file not marked beta) for all further purposes. This function was
    generated only to ease transition from grpcio<0.15.0 to grpcio>=0.15.0"""
    request_serializers = {
      ('ManagerAgent', 'create_user'): UserInfo.SerializeToString,
      ('ManagerAgent', 'execute'): RequestMessage.SerializeToString,
      ('ManagerAgent', 'refresh_resources'): UserInfo.SerializeToString,
    }
    response_deserializers = {
      ('ManagerAgent', 'create_user'): UserInfo.FromString,
      ('ManagerAgent', 'execute'): ResponseMessage.FromString,
      ('ManagerAgent', 'refresh_resources'): ResponseMessage.FromString,
    }
    cardinalities = {
      'create_user': cardinality.Cardinality.UNARY_UNARY,
      'execute': cardinality.Cardinality.UNARY_UNARY,
      'refresh_resources': cardinality.Cardinality.UNARY_UNARY,
    }
    stub_options = beta_implementations.stub_options(host=host, metadata_transformer=metadata_transformer, request_serializers=request_serializers, response_deserializers=response_deserializers, thread_pool=pool, thread_pool_size=pool_size)
    return beta_implementations.dynamic_stub(channel, 'ManagerAgent', cardinalities, options=stub_options)
except ImportError:
  pass
# @@protoc_insertion_point(module_scope)



import time
from concurrent import futures

import grpc

from eu.softfire.tub.core import CoreManagers
from eu.softfire.tub.core.CoreManagers import list_resources
from eu.softfire.tub.entities.entities import ManagerEndpoint, ResourceMetadata
from eu.softfire.tub.entities.repositories import save, find, delete, find_by_element_value
from eu.softfire.tub.messaging.grpc import messages_pb2
from eu.softfire.tub.messaging.grpc import messages_pb2_grpc
from eu.softfire.tub.utils.utils import get_logger, get_config, get_mapping_managers

logger = get_logger('eu.softfire.tub.messaging')
_ONE_DAY_IN_SECONDS = 60 * 60 * 24


def receive_forever():
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=int(get_config('system', 'server_threads', 5))))
    messages_pb2_grpc.add_RegistrationServiceServicer_to_server(RegistrationAgent(), server)
    binding = '[::]:%s' % get_config('messaging', 'bind_port', 50051)
    logger.info("Binding rpc registration server to: %s" % binding)
    server.add_insecure_port(binding)
    server.start()
    try:
        while True:
            time.sleep(_ONE_DAY_IN_SECONDS)
    except KeyboardInterrupt:
        logger.debug("Stopping server")
        server.stop(True)


def unregister_endpoint(manager_endpoint_name: str) -> bool:
    deleted = False
    for manager_endpoint in find(ManagerEndpoint):
        if manager_endpoint.name == manager_endpoint_name:
            for resource_type in get_mapping_managers().get(manager_endpoint.name):
                for rm in [rm for rm in find(ResourceMetadata) if rm.node_type.lower() == resource_type.lower()]:
                    delete(rm)
            delete(manager_endpoint)
            deleted = True
    return deleted


class RegistrationAgent(messages_pb2_grpc.RegistrationServiceServicer):
    def update_status(self, request, context):
        # logger.debug("Received request: %s" % request)
        username = request.username
        manager_name = request.manager_name
        resources = request.resources
        if username and manager_name and resources and len(resources):
            CoreManagers.update_experiment(username, manager_name, resources)
        response_message = messages_pb2.ResponseMessage()
        response_message.result = 0
        return response_message

    def unregister(self, request, context):
        logger.info("unregistering %s" % request.name)
        deleted = unregister_endpoint(request.name)
        if deleted:
            return messages_pb2.ResponseMessage(result=0)
        else:
            return messages_pb2.ResponseMessage(result=1, error_message="manager endpoint not found")

    def register(self, request, context):
        logger.info("registering %s" % request.name)
        old_managers = find_by_element_value(ManagerEndpoint, ManagerEndpoint.name, request.name)
        for old_man in old_managers:
            delete(old_man)
            logger.debug("Removed old manager endpoint: %s:%s" % (old_man.name, old_man.endpoint))
        manager_endpoint = ManagerEndpoint()
        manager_endpoint.name = request.name
        manager_endpoint.endpoint = request.endpoint
        save(manager_endpoint, ManagerEndpoint)
        list_resources()
        response_message = messages_pb2.ResponseMessage()
        response_message.result = 0
        return response_message

    def __init__(self):
        self.stop = False

    def stop(self):
        self.stop = True



from concurrent.futures import ThreadPoolExecutor, TimeoutError

from eu.softfire.tub.api import Api as api
from eu.softfire.tub.entities.repositories import drop_tables
from eu.softfire.tub.main import configuration
from eu.softfire.tub.messaging.MessagingAgent import receive_forever
from eu.softfire.tub.utils.utils import get_logger, get_config

logger = get_logger(__name__)
executor = ThreadPoolExecutor(3)
threads = []


def _setup():
    """
    Starts gRPC server in asyncio loop
    :return: loop and executor
    """
    global executor, threads
    logger.info("Starting Experiment Manager.")
    threads.append(executor.submit(receive_forever))


_setup()
application = api.app


def start_app():
    """
        Start the ExperimentManager as application
    """
    global executor, threads
    threads.append(executor.submit(api.start_listening))
    config_t = configuration.init_sys()
    cancel = False
    while True:
        for t in threads:
            try:
                if not cancel:
                    t.result(timeout=3)
                else:
                    if t.running():
                        t.cancel()
            except TimeoutError:
                continue
            except KeyboardInterrupt:
                logger.info("received ctrl-c, shutting down...")
                cancel = True
                if get_config('database', 'drop_on_exit', False).lower() == 'true':
                    drop_tables()
                configuration.stop.set()
                config_t.join()



from enum import Enum

from sqlalchemy import Column, Integer, String, PickleType, ForeignKey, Date
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship

Base = declarative_base()


class ManagerEndpoint(Base):
    __tablename__ = "manager_endpoint"

    id = Column(Integer, primary_key=True)
    name = Column(String(250), nullable=False, unique=True)
    endpoint = Column(String(250), nullable=False)


class Experimenter(Base):
    __tablename__ = "experimenter"

    id = Column(Integer, primary_key=True)
    username = Column(String(250), nullable=False)
    role = Column(String(250), nullable=False)
    ob_project_id = Column(String(250), nullable=False)
    password = Column(String(250), nullable=False)
    testbed_tenants = Column(PickleType)


class Experiment(Base):
    __tablename__ = "experiment"

    # id = Column(Integer, primary_key=True)
    id = Column(String(250), primary_key=True)
    tident = Column(String(250), primary_key=False)
    name = Column(String(250), nullable=False)
    username = Column(String(250), nullable=False)
    resources = relationship("UsedResource", cascade="all")

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __cmp__(self, other):
        return self.__dict__ == other.__dict__


class UsedResource(Base):
    __tablename__ = "used_resource"
    id = Column(Integer, primary_key=True)
    name = Column(String(250), nullable=False, unique=True)
    resource_id = Column(String(250), nullable=False, unique=False)
    parent_id = Column(String(250), ForeignKey('experiment.id'))
    status = Column(Integer, nullable=False)
    value = Column(String(7500), nullable=True)
    start_date = Column(Date, nullable=False)
    end_date = Column(Date, nullable=False)
    node_type = Column(String(250), unique=False, nullable=False)


class ResourceMetadata(Base):
    __tablename__ = "resource_metadata"
    __mapper_args__ = {
        'confirm_deleted_rows': False
    }
    # id = Column(Integer, primary_key=True)
    id = Column(String(250), primary_key=True)
    user = Column(String(250), nullable=True)
    node_type = Column(String(250), unique=False, nullable=False)
    cardinality = Column(Integer, nullable=False)
    description = Column(String(2500), nullable=False)
    testbed = Column(String(250), unique=False, nullable=True)
    properties = Column(PickleType) # used for specific properties of the resources (e.g. csar file name)


class ResourceStatus(Enum):
    VALIDATING = 0
    RESERVED = 1
    DEPLOYED = 2
    TERMINATED = 3
    ERROR = 4

    @staticmethod
    def from_int_to_enum(num):
        for val in ResourceStatus:
            if val.value == num:
                return val



import configparser
import json
import logging
import logging.config
import os
import sys
from threading import Thread, Event

from eu.softfire.tub.utils.static_config import CONFIG_FILE_PATH


def get_logger(name):
    logging.config.fileConfig(CONFIG_FILE_PATH)
    return logging.getLogger(name)


def get_config_parser():
    """
    Get the ConfigParser object containing the system configurations

    :return: ConfigParser object containing the system configurations
    """
    config = configparser.ConfigParser()
    if os.path.exists(CONFIG_FILE_PATH) and os.path.isfile(CONFIG_FILE_PATH):
        config.read(CONFIG_FILE_PATH)
        return config
    else:
        logging.error("Config file not found, create %s" % CONFIG_FILE_PATH)
        exit(1)


def get_config(section, key, default=None):
    config = get_config_parser()
    if default is None:
        return config.get(section=section, option=key)
    try:
        return config.get(section=section, option=key)
    except configparser.NoOptionError:
        return default
    except configparser.NoSectionError:
        return default


def get_mapping_managers():
    """{
    'sdn-manager': [
        'SdnResource'
    ],
    'nfv-manager': [
        'NfvResource'
    ],
    'monitoring-manager': [
        'MonitoringResource'
    ],
    'security-manager': [
        'SecurityResource'
    ],
    'physical-device-manager': [
        'PhysicalResource'
    ]
}"""

    with open(get_config('system', 'mapping-manager-file', '/etc/softfire/mapping-managers.json')) as f:
        return json.loads(f.read())


class ExceptionHandlerThread(Thread):
    def __init__(self, group=None, target=None, name=None, args=(), kwargs=None, *, daemon=None):
        if sys.version_info > (3, 0):
            super().__init__(group, target, name, args, kwargs, daemon=daemon)
        else:
            super(self.__class__, self).__init__(group, target, name, args, kwargs, daemon=daemon)
        self.exception = None

    def run(self):
        try:
            if sys.version_info > (3, 0):
                super().run()
            else:
                super(self.__class__, self).run()
        except Exception as e:
            self.exception = e


class TimerTerminationThread(Thread):
    def __init__(self, days: int, func, *args):
        super().__init__()
        self.event = Event()
        self.days = days
        self.function = func
        self.args = args

    def run(self):
        # while not self.event.wait(self.days):
        while not self.event.wait(86400 * self.days):
            self.function(*self.args[0])
            if not self.event.is_set():
                self.event.set()

    def stop(self):
        self.event.set()



import logging
import threading
from contextlib import contextmanager

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.pool import StaticPool

from eu.softfire.tub.entities import entities
from eu.softfire.tub.entities.entities import Base
from eu.softfire.tub.messaging.grpc import messages_pb2
from eu.softfire.tub.utils.utils import get_config, get_logger

logger = get_logger('eu.softfire.tub.repository')

lock = threading.RLock()

db_url = get_config('database', 'url', "sqlite:////tmp/experiment-manager.db")
if db_url.startswith("sqlite:"):
    engine = create_engine(db_url, poolclass=StaticPool, connect_args={'check_same_thread': False})
else:
    engine = create_engine(db_url)
debug_echo = (logger.getEffectiveLevel() == logging.DEBUG) and get_config('database', 'show_sql',
                                                                          False).lower() == 'true'
engine.echo = debug_echo
Base.metadata.create_all(engine)
session_factory = sessionmaker(bind=engine)
_session = scoped_session(session_factory)
session = _session()


@contextmanager
def get_db_session():
    with lock:
        with session.no_autoflush:
            yield session


def rollback():
    with get_db_session() as se:
        se.rollback()


def save(entity, _clazz=None):
    if _clazz:
        if hasattr(entity, 'id'):  # usually id is None so this method acs as normal save
            _id = entity.id
        else:
            _id = entity.name
        try:
            if _id:
                found = find(_clazz, _id)
                if isinstance(found, list):
                    for e in found:
                        delete(e)
                else:
                    if found:
                        delete(found)
        except NoResultFound:
            pass

    with get_db_session() as se:
        se.add(entity)
        se.commit()


def delete(entity):
    with get_db_session() as se:
        se.delete(entity)
        se.commit()


def find(_clazz, _id=None):
    with get_db_session() as se:
        if _id is None:
            res = se.query(_clazz).all()
        else:
            res = se.query(_clazz).filter(_clazz.id == _id).one()
        se.commit()
    return res


def drop_tables():
    Base.metadata.drop_all(engine)


def find_by_element_value(_clazz, element, value):
    with get_db_session() as se:
        res = se.query(_clazz).filter(element == value).all()
        se.commit()
    return res


def get_user_info(username):
    for ex in find(entities.Experimenter):
        if ex.username == username:
            result = messages_pb2.UserInfo()
            # result.id = ex.id
            result.name = ex.username
            result.password = ex.password
            result.ob_project_id = ex.ob_project_id
            for k, v in ex.testbed_tenants.items():
                result.testbed_tenants[k] = v
            return result


import json
import os
import traceback
import zipfile
from concurrent.futures import ThreadPoolExecutor
from datetime import timedelta, datetime

import dateparser
import grpc
import yaml
from bottle import FileUpload
from grpc._channel import _Rendezvous
from toscaparser.tosca_template import ToscaTemplate

from eu.softfire.tub.core.calendar import CalendarManager
from eu.softfire.tub.entities import entities
from eu.softfire.tub.entities.entities import UsedResource, ManagerEndpoint, ResourceMetadata, Experimenter, \
    ResourceStatus
from eu.softfire.tub.entities.repositories import save, find, delete, get_user_info, find_by_element_value
from eu.softfire.tub.exceptions.exceptions import ExperimentValidationError, ManagerNotFound, RpcFailedCall, \
    ResourceNotFound, ExperimentNotFound
from eu.softfire.tub.messaging.grpc import messages_pb2_grpc, messages_pb2
from eu.softfire.tub.utils.utils import get_logger, ExceptionHandlerThread, TimerTerminationThread, get_config, \
    get_mapping_managers

logger = get_logger('eu.softfire.tub.core')

REFRESH_RES_NODE_TYPES = [
    "NfvImage"
]

TESTBED_MAPPING = {
    'fokus': messages_pb2.FOKUS,
    'fokus-dev': messages_pb2.FOKUS_DEV,
    'ericsson': messages_pb2.ERICSSON,
    'ericsson-dev': messages_pb2.ERICSSON_DEV,
    'surrey': messages_pb2.SURREY,
    'surrey-dev': messages_pb2.SURREY_DEV,
    'ads': messages_pb2.ADS,
    'ads-dev': messages_pb2.ADS_DEV,
    'dt': messages_pb2.DT,
    'dt-dev': messages_pb2.DT_DEV,
    'any': messages_pb2.ANY,
}

MANAGERS_CREATE_USER = ['nfv-manager', 'sdn-manager']


def _get_testbed_string(testbed_id):
    for k, v in TESTBED_MAPPING.items():
        if v == testbed_id:
            return k
    raise KeyError


def create_user(username, password, role='experimenter'):
    experimenter = Experimenter()
    experimenter.username = username
    experimenter.password = password
    experimenter.role = role

    user_info = messages_pb2.UserInfo(name=username, password=password)
    for man in MANAGERS_CREATE_USER:
        try:
            user_info = get_stub_from_manager_name(man).create_user(user_info)
            logger.info("Manager %s create user finished, new UserInfo: %s" % (man, user_info))
        except ManagerNotFound:
            traceback.print_exc()
            logger.error("one of the manager is not register and need to create user")
    logger.info("Created user, project and tenant on the NFV Resource Manager")

    experimenter.testbed_tenants = {}
    experimenter.ob_project_id = user_info.ob_project_id

    for k, v in user_info.testbed_tenants.items():
        experimenter.testbed_tenants[k] = v

    save(experimenter)
    logger.info("Stored new experimenter: %s" % experimenter.username)


def delete_user(username):
    for experimenter in find(Experimenter):
        if experimenter.username == username:
            delete(experimenter)


def create_user_info(username, password, role):
    create_user(username, password, role)


def _start_termination_thread_for_res(res):
    res_end_date = res.end_date
    logger.debug("Type of res.end: %s" % type(res_end_date))
    days = (res_end_date - datetime.date(datetime.today()).today()).days
    thread = TimerTerminationThread(abs(days), _terminate_expired_resource, [res])
    logger.debug("Days to the end of experiement: %s" % days)
    logger.debug("Starting thread checking resource: %s " % res.id)
    # thread = TimerTerminationThread(5, _terminate_expired_resource, [res])
    thread.start()


def add_resource(username, id, node_type, cardinality, description, testbed, file=None):
    """
    Creates a new ResourceMetadata object and stores it in the database.
    If the file parameter is not None, it is expected to be of type FileUpload from the bottle module or str.
    In the former case the file is stored where the available NSDs reside (/etc/softfire/experiment-nsd-csar).
    In the latter case the file parameter should contain the file name so it can be added to the ResourceMetadata.

    :param id:
    :param node_type:
    :param cardinality:
    :param description:
    :param testbed:
    :param file:
    :return: the ResourceMetadata ID
    """
    logger.debug('Add new resource')
    resource_metadata = ResourceMetadata()
    resource_metadata.user = username
    resource_metadata.id = id
    resource_metadata.description = description
    resource_metadata.cardinality = cardinality
    resource_metadata.node_type = node_type
    resource_metadata.testbed = testbed
    resource_metadata.properties = {}

    if file:
        if isinstance(file, FileUpload):  # method call comes directly from api
            logger.debug('File is provided')
            nsd_csar_location = get_config('system', 'temp-csar-location',
                                           '/etc/softfire/experiment-nsd-csar').rstrip('/')
            nsd_csar_location = '{}/{}'.format(nsd_csar_location, username)
            if not os.path.exists(nsd_csar_location):
                os.makedirs(nsd_csar_location)
            logger.debug('Save file as {}/{}'.format(nsd_csar_location, '%s.csar' % file.name))
            file.save('{}/{}'.format(nsd_csar_location, '%s.csar' % file.name), overwrite=True)
            resource_metadata.properties['nsd_file_name'] = file.name
        elif isinstance(file, str):  # only the file name is provided
            logger.debug('File name is provided')
            resource_metadata.properties['nsd_file_name'] = file

    save(resource_metadata, ResourceMetadata)
    logger.debug('Saved ResourceMetadata with ID: %s' % id)
    return resource_metadata.id


class Experiment(object):
    END_DATE = 'end-date'
    START_DATE = 'start-date'

    def __init__(self, file_path, username):
        self.username = username
        self.file_path = file_path
        zf = zipfile.ZipFile(self.file_path)
        for info in zf.infolist():
            filename = info.filename
            logger.debug("Filename: %s" % filename)
            if filename.endswith(".yaml") and filename.startswith("Definitions/"):
                logger.debug("parsing %s..." % filename)
                try:
                    tpl = ToscaTemplate(yaml_dict_tpl=yaml.load(zf.read(filename)))
                except Exception as e:
                    raise ExperimentValidationError(e)
                for node in tpl.nodetemplates:
                    logger.debug("Found node: %s of type %s with properties: %s" % (
                        node.name, node.type, list(node.get_properties().keys())))
                self.topology_template = tpl
            if filename == 'TOSCA-Metadata/Metadata.yaml':
                metadata = yaml.load(zf.read(filename))
                self.name = metadata.get('name')
                # 12/12/12 10:55
                self.start_date = self.get_start_date(metadata)
                # 12/12/12 11:55
                self.end_date = self.get_end_date(metadata)
                self.duration = self.end_date - self.start_date
                logger.debug("Experiment duration %s" % self.duration)
            if filename.startswith("Files/") and filename.endswith('.csar'):

                experiment_nsd_csar_location = get_config('system', 'temp-csar-location',
                                                          '/etc/softfire/experiment-nsd-csar')

                experiment_nsd_csar_location = '{}/{}'.format(experiment_nsd_csar_location.rstrip('/'), username)
                if not os.path.exists(experiment_nsd_csar_location):
                    os.makedirs(experiment_nsd_csar_location)
                data = zf.read(filename)

                nsd_file_location = "%s/%s" % (
                    experiment_nsd_csar_location,
                    filename.split('/')[-1]
                )
                with open(nsd_file_location, 'wb+') as f:
                    f.write(data)
        temp_ids = []
        for node in self.topology_template.nodetemplates:
            if node.type == 'NfvResource':
                file_name = node.get_properties().get('file_name')
                if file_name:
                    file_name = file_name.value
                if file_name and file_name.startswith("Files/") and file_name.endswith(".csar"):
                    real_file_name = file_name[6:]
                    tmp_file_location = '{}/{}/{}'.format(get_config(
                        'system', 'temp-csar-location',
                        '/etc/softfire/experiment-nsd-csar'
                    ).rstrip('/'), username, real_file_name)
                    # get the description
                    zf = zipfile.ZipFile(tmp_file_location)
                    yaml_file = zf.read('tosca-metadata/Metadata.yaml')
                    if yaml_file:
                        yaml_content = yaml.load(yaml_file)
                        description = yaml_content.get('description')
                    else:
                        description = 'No description available'
                    testbeds = node.get_properties().get('testbeds').value
                    temp_ids.append(add_resource(
                        username,
                        node.get_properties().get('resource_id').value,
                        "NfvResource",
                        -1,
                        description,
                        list(testbeds.keys())[0],
                        real_file_name
                    ))
        try:
            self._validate()
        except Exception as e:
            for id in temp_ids:
                delete(find(ResourceMetadata, _id=id))
            raise ExperimentValidationError(e.args)

        exp = entities.Experiment()
        exp.id = "%s_%s" % (self.username, self.name)
        exp.username = self.username
        exp.name = self.name
        exp.start_date = self.start_date
        exp.end_date = self.end_date
        exp.resources = []

        for node in self.topology_template.nodetemplates:
            exp.resources.append(self._get_used_resource_by_node(node))

        logger.info("Saving experiment %s" % exp.name)
        element_value = find_by_element_value(entities.Experiment, entities.Experiment.username, self.username)
        if len(element_value):
            raise ExperimentValidationError("You cannot have two experiments at the same time!")
        save(exp)
        self.experiment = exp
        for res in exp.resources:
            _start_termination_thread_for_res(res)

    @classmethod
    def get_end_date(cls, metadata):
        try:
            return dateparser.parse(metadata.get(cls.END_DATE), settings={'DATE_ORDER': 'YMD'})
        except ValueError:
            raise ExperimentValidationError("Unknown end date format")

    @classmethod
    def get_start_date(cls, metadata):
        try:
            date_string = metadata.get(cls.START_DATE)
            return dateparser.parse(date_string, settings={'DATE_ORDER': 'YMD'})
        except ValueError:
            raise ExperimentValidationError("Unknown start date format")

    def get_topology(self):
        return self.topology_template

    def get_nodes(self, manager_name):
        nodes = []
        for node in self.topology_template.nodetemplates:
            if node.type in get_mapping_managers()[manager_name]:
                nodes.append(node)
        return nodes

    def _validate(self):
        if not self.name:
            raise ExperimentValidationError("Name must not be empty")
        if self.start_date is None or self.end_date is None:
            raise ExperimentValidationError("Not able to parse start or end date")
        if self.duration <= timedelta(0, 1, 0):
            raise ExperimentValidationError("Duration too short, modify start and end date")

        resource_ids = [rm.id for rm in find(ResourceMetadata)]
        threads = []
        for node in self.topology_template.nodetemplates:
            resource_id_ = node.get_properties()["resource_id"].value
            if (resource_id_ not in resource_ids) and (node.type != 'NfvResource'):
                raise ExperimentValidationError("resource id %s not allowed" % resource_id_)

            resource_metadata = find(ResourceMetadata, resource_id_)
            thread = ExceptionHandlerThread(target=_validate_resource, args=[node, self.username, resource_metadata])
            threads.append(thread)
            thread.start()
        for t in threads:
            t.join()
            if t.exception:
                raise t.exception

    def reserve(self):
        used_resources = []
        for node in self.topology_template.nodetemplates:
            used_resource = _get_used_resource_from_node(node, self.username)
            used_resources.append(used_resource)
            CalendarManager.check_availability_for_node(used_resource)
        try:
            CalendarManager.check_overlapping_resources(used_resources)
        except ExperimentValidationError as e:
            delete(self.experiment)
            raise e
        # all node have been granted

        for us in self.experiment.resources:
            us.status = ResourceStatus.RESERVED.value

        save(self.experiment)

    def _get_used_resource_by_node(self, node):
        resource = UsedResource()
        resource.name = node.name
        resource.node_type = node.type
        resource.value = json.dumps(node.entity_tpl)
        resource.resource_id = node.get_properties().get('resource_id').value
        resource.status = ResourceStatus.VALIDATING.value
        if node.get_properties().get(self.START_DATE):
            resource.start_date = Experiment.get_start_date(node.get_properties())
        else:
            resource.start_date = self.start_date

        if node.get_properties().get(self.END_DATE):
            resource.end_date = Experiment.get_end_date(node.get_properties())
        else:
            resource.end_date = self.end_date
        return resource


def _release_used_resource(res: entities.UsedResource):
    exp = find(entities.Experiment, _id=res.parent_id)
    user_info = get_user_info(exp.username)
    for man, node_types in get_mapping_managers().items():
        if res.node_type in node_types:
            response = get_stub_from_manager_name(man).execute(
                messages_pb2.RequestMessage(method=messages_pb2.RELEASE_RESOURCES,
                                            payload=res.value,
                                            user_info=user_info))
            if response.result != 0:
                logger.error("release resources returned %d: %s" % (response.result, response.error_message))
                raise RpcFailedCall(
                    "provide resources returned %d: %s" % (response.result, response.error_message))


def _terminate_expired_resource(res: entities.UsedResource):
    resource_to_delete = find(entities.UsedResource, _id=res.id)
    exp_id = resource_to_delete.parent_id
    if resource_to_delete and res == resource_to_delete:
        _release_used_resource(res)
        delete(resource_to_delete)
    else:
        logger.debug("The initial experiment was removed")
        return
    exp = find(entities.Experiment, _id=exp_id)
    if not len(exp.resources):
        delete(exp)


def _get_used_resource_from_node(node, username):
    for e in find(entities.Experiment):
        if e.username == username:
            for ur in find(UsedResource):
                if ur.parent_id == e.id and ur.name == node.name:
                    return ur

    raise ResourceNotFound('Resource with name %s  for user %s was not found' % (node.name, username))


def get_stub_from_manager_name(manager_name):
    for manager_endpoint in find(ManagerEndpoint):
        if manager_endpoint.name == manager_name:
            logger.debug("Getting stub for manager: %s" % manager_name)
            return get_stub_from_manager_endpoint(manager_endpoint)
    raise ManagerNotFound("No manager found for name %s" % manager_name)


def get_stub_from_manager_endpoint(manager_endpoint):
    endpoint = manager_endpoint.endpoint
    logger.debug("looking for endpoint %s" % endpoint)
    channel = grpc.insecure_channel(endpoint)
    return messages_pb2_grpc.ManagerAgentStub(channel)


def _validate_resource(node, username, request_metadata):
    for manager_endpoint in find(ManagerEndpoint):
        if node.type in get_mapping_managers().get(manager_endpoint.name):
            if request_metadata and request_metadata.properties and request_metadata.properties.get('nsd_file_name'):
                nsd_file_name = request_metadata.properties.get('nsd_file_name')
                node.entity_tpl.get('properties')['file_name'] = 'Files/%s' % nsd_file_name
            request_message = messages_pb2.RequestMessage(method=messages_pb2.VALIDATE_RESOURCES,
                                                          payload=yaml.dump(node.entity_tpl),
                                                          user_info=get_user_info(username))
            try:
                response = get_stub_from_manager_endpoint(manager_endpoint).execute(request_message)
            except Exception as e:
                if hasattr(e, 'message'):
                    raise RpcFailedCall(e.message)
                raise RpcFailedCall(e.args)

            if response.result == messages_pb2.ERROR:
                raise RpcFailedCall(response.error_message)
            return

    raise ManagerNotFound("manager handling resource %s was not found" % node.type)


def list_resources(username=None, manager_name=None):
    managers = []
    if manager_name is None:
        for man in find(ManagerEndpoint):
            managers.append(man.name)
    else:
        managers.append(manager_name)

    result = []
    # List resources can be done in parallel
    max_workers = len(managers)
    tpe = ThreadPoolExecutor(max_workers=max_workers)
    threads = []
    for manager in managers:
        threads.append(tpe.submit(_execute_rpc_list_res, manager))
    for t in threads:
        result.extend(t.result())

    logger.debug("Saving %d resources" % len(result))

    for rm in find(entities.ResourceMetadata):
        for rm_to_del in result:
            if rm.id == rm_to_del.resource_id:
                delete(rm)

    for rm in result:
        if username is None or rm.user == username:
            resource_metadata = ResourceMetadata()
            if username:
                resource_metadata.user = username
            resource_metadata.id = rm.resource_id
            resource_metadata.description = rm.description
            resource_metadata.cardinality = rm.cardinality
            resource_metadata.node_type = rm.node_type
            resource_metadata.testbed = _get_testbed_string(rm.testbed)
            save(resource_metadata, ResourceMetadata)

    return result


def _execute_rpc_list_res(manager):
    stub = get_stub_from_manager_name(manager)
    request_message = messages_pb2.RequestMessage(method=messages_pb2.LIST_RESOURCES, payload='', user_info=None)
    response = stub.execute(request_message)
    if response.result != 0:
        logger.error("list resources returned %d: %s" % (response.result, response.error_message))
        raise RpcFailedCall("list resources returned %d: %s" % (response.result, response.error_message))
    return response.list_resource.resources


def provide_resources(username):
    experiments_to_deploy = find_by_element_value(entities.Experiment, entities.Experiment.username, username)
    if len(experiments_to_deploy) == 0:
        logger.error("No experiment to be deleted....")
        raise ExperimentNotFound("No experiment to be deployed....")
    experiment_to_deploy = experiments_to_deploy[0]

    user_info = get_user_info(username)
    if hasattr(user_info, 'name'):
        un = user_info.name
    else:
        un = user_info.username
    logger.debug("Received deploy resources from user %s" % un)
    logger.debug("Received deploy resources %s" % experiment_to_deploy.name)

    involved_managers = [man_name for ur in experiment_to_deploy.resources for man_name, node_types in
                         get_mapping_managers().items() if ur.node_type in node_types]

    manager_ordered = get_config('system', 'deployment-order', '').split(';')
    remaining_managers = set(list(get_mapping_managers().keys())) - set(manager_ordered)
    for manager_name in manager_ordered:
        if manager_name in involved_managers:
            _provide_all_resources_for_manager(experiment_to_deploy, manager_name, user_info)

    for manager_name in remaining_managers:
        if manager_name in involved_managers:
            _provide_all_resources_for_manager(experiment_to_deploy, manager_name, user_info)


def _provide_all_resources_for_manager(experiment_to_deploy, manager_name, user_info):
    stub = get_stub_from_manager_name(manager_name)
    for ur_to_deploy in experiment_to_deploy.resources:
        node_types = get_mapping_managers().get(manager_name)
        if ur_to_deploy.node_type in node_types:
            response = stub.execute(messages_pb2.RequestMessage(method=messages_pb2.PROVIDE_RESOURCES,
                                                                payload=ur_to_deploy.value,
                                                                user_info=user_info))
            for ur in experiment_to_deploy.resources:
                if ur.id == ur_to_deploy.id:
                    if response.result < 0:
                        logger.error(
                            "provide resources returned %d: %s" % (response.result, response.error_message))
                        # raise RpcFailedCall(
                        #     "provide resources returned %d: %s" % (response.result, response.error_message))
                        ur.status = ResourceStatus.ERROR.value
                        continue
                    ur.value = ""
                    for res in response.provide_resource.resources:
                        logger.debug("Received: %s" % str(res.content))
                        ur.value += res.content
                    ur.status = ResourceStatus.DEPLOYED.value


def release_resources(username):
    experiments_to_delete = find_by_element_value(entities.Experiment, entities.Experiment.username, username)
    if len(experiments_to_delete) == 0:
        logger.error("No experiment to be deleted....")
        raise ExperimentNotFound("No experiment to be deleted....")
    experiment_to_delete = experiments_to_delete[0]
    _release_all_experiment_resources(experiment_to_delete, username)

    logger.info("deleting %s" % experiment_to_delete.name)
    delete(experiment_to_delete)


def _release_all_experiment_resources(experiment_to_delete, username):
    user_info = get_user_info(username)
    used_resources = experiment_to_delete.resources
    managers = [man_name
                for man_name, node_types in get_mapping_managers().items()
                for ur in used_resources
                if ur.node_type in node_types]
    threads = []
    for manager_name in managers:
        thread = ExceptionHandlerThread(target=_release_resource_for_manager,
                                        args=[manager_name, used_resources, user_info])
        threads.append(thread)
        thread.start()
    for t in threads:
        t.join()
        if t.exception:
            raise t.exception


def _release_resource_for_manager(manager_name, used_resources, user_info):
    stub = get_stub_from_manager_name(manager_name)
    try:

        for ur in used_resources:
            if ur.node_type in get_mapping_managers().get(manager_name):
                if ur.value:
                    response = stub.execute(messages_pb2.RequestMessage(method=messages_pb2.RELEASE_RESOURCES,
                                                                        payload=ur.value,
                                                                        user_info=user_info))
                    if response.result != 0:
                        logger.error("release resources returned %d: %s" % (response.result, response.error_message))
                        raise RpcFailedCall(
                            "provide resources returned %d: %s" % (response.result, response.error_message))
                    for u in [u for u in used_resources if u.node_type in get_mapping_managers().get(manager_name)]:
                        logger.info("deleting %s" % u.name)
                        delete(u)
    except _Rendezvous:
        traceback.print_exc()
        logger.error("Exception while calling gRPC, maybe %s Manager is down?" % manager_name)
        raise RpcFailedCall("Exception while calling gRPC, maybe %s Manager is down?" % manager_name)


def get_other_resources():
    images = []
    networks = []
    flavours = []
    for rm in find(ResourceMetadata):
        if rm.node_type == 'NfvImage':
            tmp = {
                'resource_id': rm.id,
                'node_type': rm.node_type,
                'testbed': rm.testbed,
                'description': rm.description
            }
            if rm.cardinality < 0:
                tmp['cardinality'] = 'infinite'
            else:
                tmp['cardinality'] = rm.cardinality

            images.append(tmp)
        if rm.node_type == 'NfvFlavor':
            tmp = {
                'resource_id': rm.id,
                'node_type': rm.node_type,
                'testbed': rm.testbed,
                'description': rm.description
            }
            if rm.cardinality < 0:
                tmp['cardinality'] = 'infinite'
            else:
                tmp['cardinality'] = rm.cardinality

            flavours.append(tmp)
        if rm.node_type == 'NfvNetwork':
            tmp = {
                'resource_id': rm.id,
                'node_type': rm.node_type,
                'testbed': rm.testbed,
                'description': rm.description
            }
            if rm.cardinality < 0:
                tmp['cardinality'] = 'infinite'
            else:
                tmp['cardinality'] = rm.cardinality

            networks.append(tmp)

    return images, networks, flavours


def get_experiment_dict(username):
    res = []
    exp_name = "Your Experiment"
    for ex in find(entities.Experiment):
        if ex.username == username:
            exp_name += ": %s" % ex.name
            for ur in ex.resources:
                tmp = {
                    'resource_id': ur.resource_id,
                    'status': ResourceStatus.from_int_to_enum(ur.status).name,
                    'value': ur.value
                }
                res.append(tmp)
    return exp_name, res


def get_resources_dict(username=None):
    res = []
    for rm in find(ResourceMetadata):
        if rm.node_type != 'NfvImage' and rm.node_type != 'NfvNetwork' and rm.node_type != 'NfvFlavor':
            if username:
                if username == rm.user:
                    res.append(_get_resource_dict_from_rm(rm))
            else:
                res.append(_get_resource_dict_from_rm(rm))
    return res


def _get_resource_dict_from_rm(rm):
    tmp = {
        'resource_id': rm.id,
        'node_type': rm.node_type,
        'description': rm.description,
        'testbed': rm.testbed,
    }
    if rm.cardinality < 0:
        tmp['cardinality'] = 'infinite'
    else:
        tmp['cardinality'] = rm.cardinality
    return tmp


def refresh_resources(username, manager_name=None):
    managers = []
    if manager_name is None:
        for man in find(ManagerEndpoint):
            managers.append(man.name)
    else:
        managers.append(manager_name)

    user_info = get_user_info(username)
    result = []
    if user_info:
        for manager in managers:
            stub = get_stub_from_manager_name(manager)
            request_message = user_info
            try:
                response = stub.refresh_resources(request_message)
            except _Rendezvous:
                traceback.print_exc()
                logger.error("Exception while calling gRPC, maybe %s Manager is down?" % manager)
                raise RpcFailedCall("Exception while calling gRPC, maybe %s Manager is down?" % manager)
            if response.result != 0:
                logger.error("list resources returned %d: %s" % (response.result, response.error_message))
                raise RpcFailedCall("list resources returned %d: %s" % (response.result, response.error_message))
            result.extend(response.list_resource.resources)

    logger.debug("Saving %d resources" % len(result))

    for rm in find(entities.ResourceMetadata):
        if rm.node_type in REFRESH_RES_NODE_TYPES:
            delete(rm)

    for rm in result:
        resource_metadata = ResourceMetadata()
        if rm.resource_id and (not hasattr(rm, 'user') or rm.user == username):
            resource_metadata.id = rm.resource_id
            resource_metadata.description = rm.description
            resource_metadata.cardinality = rm.cardinality
            resource_metadata.node_type = rm.node_type
            if hasattr(rm, 'user') and rm.user:
                resource_metadata.user = rm.user
            resource_metadata.testbed = list(TESTBED_MAPPING.keys())[list(TESTBED_MAPPING.values()).index(rm.testbed)]
            save(resource_metadata, ResourceMetadata)

    return result


def get_used_resources_by_experimenter(exp_name):
    res = []
    for ur in find(UsedResource):
        experimenter = find(Experimenter, ur.parent_id)
        if experimenter is not None and experimenter.name == exp_name:
            res.append(ur)
    return res


def update_experiment(username, manager_name, resources):
    experiment = find_by_element_value(entities.Experiment, entities.Experiment.username, username)[0]
    index = 0

    for ur in experiment.resources:
        if ur.node_type in get_mapping_managers().get(manager_name):
            ur.value = json.dumps(json.loads(resources[index].content))
            index += 1


def list_managers():
    return [man.name for man in find(ManagerEndpoint)]


def list_experimenters():
    return [man.username for man in find(Experimenter)]



from eu.softfire.tub.entities.entities import UsedResource, ResourceStatus, ResourceMetadata
from eu.softfire.tub.entities.repositories import find
from eu.softfire.tub.exceptions.exceptions import ResourceAlreadyBooked, ExperimentValidationError


class CalendarManager(object):
    def __init__(self):
        pass

    @classmethod
    def check_availability_for_node(cls, used_resource):
        if not cls.check_overlapping_booking(used_resource):
            raise ResourceAlreadyBooked('Some resources where already booked, please check the calendar')

    @classmethod
    def check_overlapping_booking(cls, used_resource):
        """
        Check calendar availability of this resource

        :param used_resource: the used resource to book
         :type used_resource: UsedResource
        :return: True when availability is granted False otherwise
         :rtype: bool
        """
        resource_metadata = cls.get_metadata_from_usedresource(used_resource)
        # if the resource metadata associated has infinite cardinality return true
        if resource_metadata.cardinality <= 0:
            return True
        max_concurrent = resource_metadata.cardinality
        counter = 0
        # if not then i need to calculate the number of resource already booked for that period
        for ur in find(UsedResource):
            if ur.status == ResourceStatus.RESERVED.value:
                if ur.start_date <= used_resource.start_date <= ur.end_date:
                    counter += 1
                elif ur.start_date <= used_resource.end_date <= ur.end_date:
                    counter += 1
        if counter < max_concurrent:
            return True

        return False

    @classmethod
    def get_metadata_from_usedresource(cls, used_resource):
        return find(ResourceMetadata, _id=used_resource.resource_id)

    @classmethod
    def get_month(cls):
        result = []

        for ur in find(UsedResource):
            rm = find(ResourceMetadata, _id=ur.resource_id)
            if rm and rm.cardinality > 0:
                result.append({
                    "resource_id": ur.resource_id,
                    "start": ur.start_date,
                    "end": ur.end_date,
                })

        return result

    @classmethod
    def check_overlapping_resources(cls, used_resources: list):
        for ur_to_check in used_resources:
            counter = 1
            for other_res in used_resources:
                if other_res != ur_to_check:
                    if _time_overlap(ur_to_check.start_date, ur_to_check.end_date, other_res.start_date,
                                     other_res.end_date):
                        counter += 1
            cardinality = find(ResourceMetadata, _id=ur_to_check.resource_id).cardinality
            if 0 < cardinality < counter:
                raise ExperimentValidationError(
                    "Trying to book %d resources of resource_id %s" % (counter, ur_to_check.resource_id))


def _time_overlap(start_date_src, end_date_src, start_date_dst, end_date_dst):
    overlap = (min(end_date_src, end_date_dst) - max(start_date_src, start_date_dst))
    days = overlap.days + 1
    hours, remainder = divmod(overlap.seconds, 3600)
    minutes, seconds = divmod(remainder, 60)
    hours += days * 24
    return hours > 0 or minutes > 0, seconds > 0
